/*--[drm5.c]-------------------------------------------------------------------
 | Copyright (C) 2002, 2003 Dan A. Jackson 
 |
 | This file is part of the "clit" (Convert LIT) program. 
 |
 | Convert LIT is free software; you can redistribute it and/or modify
 | it under the terms of the GNU General Public License as published by
 | the Free Software Foundation; either version 2 of the License, or
 | (at your option) any later version.
 |
 | This program is distributed in the hope that it will be useful,
 | but WITHOUT ANY WARRANTY; without even the implied warranty of
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 | GNU General Public License for more details.
 |
 | You should have received a copy of the GNU General Public License
 | along with this program; if not, write to the Free Software
 | Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 | 
 | The GNU General Public License may also be available at the following
 | URL: http://www.gnu.org/licenses/gpl.html
*/
/* This file contains a way to slip through the DRM5 fences.  It's only 
 | a stop-gap measure, and there are always other ways.   */

/* 
 | Update for June 4th, the day Americans celebrate their freedom to write
 | software that restores fair-use rights.   
 | 
 | After a 6 month hiatus Microsoft released an update which checked that
 | the filename calling the Secure Repository DLL was actually inside the
 | manifest.  Unfortunately, ConvertLit is a terrible lying cheat of a
 | program, and tells the DLL exactly what it wants to hear.
 */


#ifdef _MSC_VER
#include <windows.h>
#include <winbase.h>
#include <winnt.h>      /* PE-Format defines */
#include <direct.h>
#endif

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "litlib.h"

static const char * license_string = "/DRMStorage/Licenses/EUL";

typedef int (*voidfunc)(void);
typedef int (__stdcall *bbfunc)(void **);
typedef int (*initfunc)(void *,int, void *);
typedef int (*setrsafunc)(void *,int, unsigned char *, unsigned long int *);
typedef int (*setsymfunc)(void *,int, int, void *, unsigned long int *);
typedef int (*decfunc)(void *,unsigned long int,int,unsigned char *, unsigned long int *, unsigned char *);

typedef struct bbtable {
    void * f1;
    void * f2;
    initfunc init;
    void * f3;
    setrsafunc setrsa;
    void * f5;
    setsymfunc setsym;
    decfunc    decrypt;
    U32         test[64];
} bbtable;

extern int jumpgate(void);

int drm5_setkey(U8 * p, int size);
int drm5_load_secure_repository(void);

int      paramcount = 0;
voidfunc trampoline_target = NULL;
unsigned long int ret_addr = 0;
void * bbobj;
bbtable * pTable;
unsigned int key_tag;
HMODULE  hDRM;


int MS_Base64Decode(wchar_t * str, unsigned char **buff)
{
    wchar_t *cp;
    char *ocp;
    int len, val, count, block, ocount;

    len = wcslen(str);
    if ((*buff = malloc((len * 3) / 4)) == NULL) return -1;
    ocp = *buff;
    count = 0;
    block = 0;
    ocount = 0;
    for (cp = str; *cp != L'\0'; cp++) {
        if ((*cp >= L'A') && (*cp <= L'Z'))
            val = *cp - L'A';
        else if ((*cp >= L'a') && (*cp <= L'z'))
            val = *cp - L'a' + 26;
        else if ((*cp >= L'0') && (*cp <= L'9'))
            val = *cp - L'0' + 52;
        else if ((*cp == L'+') || (*cp == L'!'))
            val = 62;
        else if ((*cp == L'/') || (*cp == L'*'))
            val = 63;
        else if (*cp == L'=') {
            if (count == 2) {
                *ocp++ = block >> 4;
                ocount++;
            } else {
                *ocp++ = (block >> 10);
                *ocp++ = (block >> 2) & 0xff;
                ocount += 2;
            }
            break;
        } else
            val = 1;

        if (val >= 0) {
            block = (block << 6) | val;
            if (++count == 4) {
                *ocp++ = block >> 16;
                *ocp++ = (block >> 8) & 0xff;
                *ocp++ = block & 0xff;
                ocount += 3;
                count = 0;
            }
        }
    }
    return ocount;
}

/* Stupid little fake XML parser. */

static wchar_t *find_close(wchar_t * str)
{
    while ((*str != L'\0') && (*str != L'>')) {
        if (*str == L'"') {
            if ((str = wcschr(str + 1, L'"')) == NULL)
                return NULL;
        }
        str++;
    }

    if (*str == L'\0')
        return NULL;
    else
        return str + 1;
}



wchar_t *get_element(wchar_t * tag, wchar_t * str)
{
    int len = wcslen(tag);
    wchar_t *tmptag;
    wchar_t *start, *end;
    wchar_t *rval = NULL;
    wchar_t *tmp;

    if ((tmptag = malloc((len + 4) * sizeof(wchar_t))) == NULL) {
        lit_error(ERR_R, "Memory allocation failed in get_element()"); 
        return NULL;
    }
    swprintf(tmptag, L"<%s", tag);

    while (1) {
        if ((start = wcsstr(str, tmptag)) == NULL)
            goto exit;
        if (!iswalnum(start[len + 1]))
            break;
        str = start + len + 1;
    }

    swprintf(tmptag, L"</%s>", tag);
    end = wcsstr(str, tmptag);
    if (end == NULL) {
        goto exit;
    } else {
        wchar_t *realstart = find_close(start);
        if ((realstart == NULL) || (realstart > end)) 
            goto exit;
        tmp = malloc((end - realstart + 1) * sizeof(wchar_t));
        if (tmp == NULL) {
            lit_error(ERR_R, "Memory allocation #2 failed in get_element()"); 
            return NULL;
        }
        memcpy(tmp, realstart, (end - realstart) * sizeof(wchar_t));
        tmp[end - realstart] = L'\0';
        rval = tmp;
    }
exit:
    free(tmptag);
    return rval;
}

int drm5_handle_key(lit_file * plf)
{
    U8  * pData;
    int size, r;
    wchar_t  *ebits = NULL, * value = NULL;
    unsigned char * keyData;
    int keySize, status;
    U8 * pValidation;

    status = lit_get_file(plf,license_string,&pData,&size);
    if (status) {
        lit_error(ERR_R, 
"Was unable to find an End-User-License (\"%s\") in the LIT file.", 
	license_string); 
        return status;
    }
    ebits = get_element(L"ENABLINGBITS", (wchar_t *)pData);
    if (ebits) value = get_element(L"VALUE", ebits);
    if (!value) {
        lit_error(ERR_R, 
"I couldn't find an encrypted key in the End User License file.");
        free(pData);
        return -1;
    }
    keySize = MS_Base64Decode(value, &keyData);
    r =  drm5_setkey(keyData, keySize);
    if (!r)
    {
        status = lit_get_file(plf,"/DRMStorage/ValidationStream",&pValidation,
            &size);
        if ((status) || (!pValidation) || (!size)) {
            lit_error(ERR_R,"Unable to read validation stream.");
            free(pData);
            return -1;
        }
        if (strncmp(pValidation,"MSReader",8) != 0) { r = -1; }
        if (r == -1)
        {
            int i;
            /* Cleanse the bad sections */
            for (i = 0; i < plf->num_sections; i++)
            {
                if ((plf->sections[i].data_pointer) && 
                    (plf->sections[i].data_pointer != (U8 *)-1))
                {
                    free(plf->sections[i].data_pointer);
                    plf->sections[i].data_pointer = NULL;
                }
            }
            /*
            FreeLibrary(hDRM);
            r = drm5_load_secure_repository(); */ 
            r = 0;
            __try
            {
                if (!r) r = drm5_setkey(keyData, keySize|0x10000);
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                lit_error(ERR_R,"Activation DLL doesn't like new method.");
                r = -1;
            }
        }
        if (!r)
        {
            free(pValidation);    
            status = lit_get_file(plf,"/DRMStorage/ValidationStream",
                &pValidation, &size);
            if ((status) || (!pValidation) || (!size)) {
                lit_error(ERR_R,"Unable to read validation stream.");
                free(pData);
                return -1;
            }
            if (strncmp(pValidation,"MSReader",8) != 0) { r = -1; }
        }
        if (r != 0)
        {
            lit_error(ERR_R,"Failed to decrypt this title!");
        }
    }
    free(pData);
    free(pValidation);
    return r;
}

char * get_registry_entry(HKEY hkey, char * subkey, char * value, int pad)
{
    long    r;
    char    * buff;
    U32     len, type;
    HKEY    key;

    r = RegOpenKeyExA(hkey,subkey,0,KEY_READ,&key);
    if (r) return NULL;
    r = RegQueryValueExA(key,value,NULL,NULL,NULL,&len);
    if (r) return NULL;
    buff = (char *)malloc(len + pad);
    if (!buff) {
        lit_error(ERR_R, "Unable to get memory for registry entry.");
        return NULL;
    }
    r = RegQueryValueExA(key,value,NULL,&type,buff, &len);
    if (r) { free(buff); return NULL;}
    RegCloseKey(key);
    return buff;
}

int  get_code_info(HMODULE image_base, U32   * pBase, U32   * pSize)
{
    PIMAGE_DOS_HEADER pmz;
    PIMAGE_NT_HEADERS ppe;
    PIMAGE_OPTIONAL_HEADER popthdr;

    pmz = (PIMAGE_DOS_HEADER)image_base;
    if (pmz->e_magic != IMAGE_DOS_SIGNATURE) return 0x10;

    ppe = (PIMAGE_NT_HEADERS)((BYTE *)image_base + pmz->e_lfanew);
    if (ppe->Signature != IMAGE_NT_SIGNATURE) return 0x20;
    if (ppe->FileHeader.SizeOfOptionalHeader == 0) return 0x30;
    popthdr = (PIMAGE_OPTIONAL_HEADER)&ppe->OptionalHeader;
    if (popthdr->Magic != IMAGE_NT_OPTIONAL_HDR_MAGIC) return 0x40;

    if (pBase) *pBase = popthdr->BaseOfCode;
    if (pSize) *pSize = popthdr->SizeOfCode;
    return 0;
}

int  get_imports(HMODULE image_base, U32   * pBase, U32   * pSize)
{
    PIMAGE_DOS_HEADER pmz;
    PIMAGE_NT_HEADERS ppe;
    PIMAGE_OPTIONAL_HEADER popthdr;

    pmz = (PIMAGE_DOS_HEADER)image_base;
    if (pmz->e_magic != IMAGE_DOS_SIGNATURE) return 0x10;

    ppe = (PIMAGE_NT_HEADERS)((BYTE *)image_base + pmz->e_lfanew);
    if (ppe->Signature != IMAGE_NT_SIGNATURE) return 0x20;
    if (ppe->FileHeader.SizeOfOptionalHeader == 0) return 0x30;
    popthdr = (PIMAGE_OPTIONAL_HEADER)&ppe->OptionalHeader;
    if (popthdr->Magic != IMAGE_NT_OPTIONAL_HDR_MAGIC) return 0x40;

    if (pBase) *pBase = 
        popthdr->DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
    if (pSize) *pSize = 
        popthdr->DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].Size;
    return 0;
}


char * liblist[] = { "EBOOKFX.DLL","EBRICHED.DLL",
 "DMGR.DLL","MSLS31.DLL" };

const char * sig_string = "-manifest.sig";


HMODULE hReader;

HMODULE WINAPI HackedGetModuleHandleA( char * lpModuleName)
{
    HMODULE h;

    if (!lpModuleName) { return hReader; };
    h = GetModuleHandleA(lpModuleName);
    return h;
}

char reader_path[260];

DWORD WINAPI HackedGetModuleFileNameA( HMODULE hModule, 
    LPTSTR lpFilename, DWORD nSize) 
{

    if (!hModule)
    {
        DWORD chars_copied;

        chars_copied = 0;
        if (nSize)
            chars_copied = (strncpy(lpFilename,reader_path,nSize) - lpFilename);
        if (nSize == chars_copied)
        {
            /*strncpy doesn't do this for me */
            if (nSize) lpFilename[nSize-1] = '\0';
        }
        return strlen(reader_path)+1;
    }
    else
    {
        DWORD r;

        r = GetModuleFileNameA(hModule, lpFilename, nSize);
	return r;
    } 
}

int patch_dll(void ** ptarget, void * new_data, void ** pold)
{
    BOOL b;
    DWORD OldProtect;

    if (pold) *pold = *ptarget;
     
    b = VirtualProtect(ptarget,sizeof(*ptarget),
        PAGE_EXECUTE_READWRITE, &OldProtect);
    if (!b)
    {
        lit_error(ERR_WIN32, "VirtualProtect failed.\n");
        return -1;
    }
                    
    (void)WriteProcessMemory(GetCurrentProcess(), ptarget, &new_data,
        sizeof(new_data),NULL);
    if (*ptarget != new_data) 
    {
        lit_error(ERR_WIN32,
"WriteProcessMemory couldn't patch DLL.  Why?\n");
        return -1;
    }
    b = VirtualProtect(ptarget,sizeof(*ptarget),
        OldProtect, &OldProtect);
    if (!b)
    {
        lit_error(ERR_WIN32, "VirtualProtect failed.\n");
        return -1;
    }  
    return 0;
}

int install_hacked_api(HMODULE image_base)
{
    U32     base, size;
    PIMAGE_IMPORT_DESCRIPTOR    pi;
    int     r;
    char    * lib_name;
    void    * fcnGetModuleFileNameA, *fcnExitProcess;
    void    * fcnGetModuleHandleA;
    HMODULE hLib;

    if ((r = get_imports(image_base, &base, &size)) != 0)
        return r;
    if (!size) return 0;
    pi = (PIMAGE_IMPORT_DESCRIPTOR)((U8 *)image_base + base);
    while (size)
    {
        /* Catch sentinel entry */
        if (!pi->Name) break;
        lib_name = (U8 *)image_base + pi->Name;

        if (_stricmp(lib_name,"kernel32.dll") == 0)
        {
            void    ** pfcn;

            hLib = GetModuleHandle(lib_name);
            fcnGetModuleHandleA = (void *)GetProcAddress(hLib,
                "GetModuleHandleA");
            fcnGetModuleFileNameA = (void *)GetProcAddress(hLib,
                "GetModuleFileNameA");
            fcnExitProcess = (void *)GetProcAddress(hLib,"ExitProcess");
            pfcn = (void **)((U8 *)image_base + (U32)pi->FirstThunk);
            while (*pfcn)
            {
                if (*pfcn == fcnGetModuleFileNameA)
                {
                    r = patch_dll(pfcn,&HackedGetModuleFileNameA,NULL);
                    if (r) return r;
                }
                else if (*pfcn == fcnGetModuleHandleA)
                {
                    r = patch_dll(pfcn,&HackedGetModuleHandleA,NULL);
                    if (r) return r;
                }


                pfcn++;
            }
        }
        pi++;
        size -= sizeof(IMAGE_IMPORT_DESCRIPTOR);
    }
    return 0;
}

int drm5_load_secure_repository(void)
{
    char * path;
    bbfunc  func;
    volatile initfunc hackinit;
    int     r;
    
    path = get_registry_entry(HKEY_CURRENT_USER,
        "Software\\Microsoft\\DRM","SecRepLocation",0);
    if (!path) {
        lit_error(0, "No DRM path in the registry - are you activated??");
        return -1;
    }
    hDRM = LoadLibraryExA(path, NULL, 0);
    if (!hDRM) {
	lit_error(ERR_WIN32,"Error loading DRM DLL: \"%s\" !",path);
        free(path);
        return -1;
    }
  
    r = install_hacked_api(hDRM);
    if (r) {
        lit_error(ERR_WIN32,"Error installing re-name hack.\n");
        free(path);
        return -1;
    }

    func = (bbfunc)GetProcAddress(hDRM,"DllGetBlackBox");
    if (!func) {
        lit_error(0,"Unable to find DllGetBlackBox in \"%s\".",path); 
        return -1;
    }
    free(path); 
    path = NULL;

    r = func(&bbobj);
    if ((r) || (!bbobj)) {
        lit_error(0,"DllGetBlackBox() call failed! Status=%d.",r);
        return -1;
    }
    pTable = (bbtable *)( *(bbtable **) bbobj );
    
    pTable = (bbtable *)( *(bbtable **) bbobj );

    trampoline_target = (voidfunc)pTable->init;
    hackinit = (initfunc)jumpgate;
    paramcount = 3;
    r = hackinit(bbobj, GetCurrentProcessId(), NULL);
    if (r < 0) { 
        lit_error(0,"Blackbox \"Initialize\" failed! (%d)\n", r);
        return -1;
    }
    return 0;
}

int drm5_init(void)
{
    char * path, *pathfile;
    int     r;
    unsigned int i;
    HMODULE hLib;
    unsigned char * pTrash;
    U32     code_base, code_size;
    char    * old_directory;

    path = get_registry_entry(HKEY_LOCAL_MACHINE,
        "Software\\Microsoft\\EBook", "InstallDir", 20);
    if (!path) {
        lit_error(0,"No Registry key to locate MSREADER.EXE directory! \n");
        return -1;
    }

    /* Fix search-path issue on Win9X */
    old_directory = _getcwd(NULL, 512);
    if (!old_directory) {
        lit_error(ERR_LIBC,"What, no current directory?!");
    }
    if (old_directory) {
        _chdir(path);
    }
    pathfile = path+strlen(path);
    strcat(path,"MSREADER.EXE");

    if (strlen(path) > sizeof(reader_path))
    {
        lit_error(0,"Programmer Sloppiness. "
            "reader_path is too short (%d < %d).\n",sizeof(reader_path),
            strlen(path));
        free(path);
        return -1;
    }
    memcpy(reader_path, path, strlen(path));


    hReader = LoadLibraryExA(path,NULL, DONT_RESOLVE_DLL_REFERENCES);
    if (!hReader) {
        lit_error(ERR_WIN32,"Failed to load MSREADER.EXE as a DLL!");
        if (old_directory) { _chdir(old_directory); free(old_directory); }
        return -1;
    }
    for (i = 0; i < sizeof(liblist)/sizeof(liblist[0]); i++) {
        strcpy(pathfile, liblist[i]);
        hLib = LoadLibraryExA(path, NULL, DONT_RESOLVE_DLL_REFERENCES); 
        if (!hLib) {
            lit_error(ERR_WIN32,"Unable to load \"%s\".",liblist[i]);
            if (old_directory) { _chdir(old_directory); free(old_directory); }
            return -1;
        }
    }
    if (old_directory) { _chdir(old_directory); free(old_directory); }
    old_directory = NULL;
 
    r = get_code_info(hReader, &code_base, &code_size);
    if (r) { 
        lit_error(0,
"I couldn't understand the PE Header for MSREADER.EXE (Code: %02d).", 
        r);
        return -1;
    }    
    ret_addr = 0;
    pTrash = (unsigned char *)hReader + code_base;
    for (i = 5; i < code_size; i++) {
        if ( *(pTrash + i) == 0xC3) {
            if ( (*(pTrash + i - 5) == 0xFF) ||
                 (*(pTrash + i - 5) == 0xE8) ||
                 (*(pTrash + i - 3) == 0xFF) ||
                 (*(pTrash + i - 3) == 0xE8) ) {
                ret_addr = (unsigned long int)(pTrash + i);
                break;
            }
        }
    }
    free(path);
    if (!ret_addr) {
        lit_error(0,
"I can't find a decent thunkable spot.  No crack for you!");
        return -1;
    }
    path = get_registry_entry(HKEY_LOCAL_MACHINE,
        "Software\\Microsoft\\EBook", "ClearTypeLocation", 0);
    if (!path) {
        lit_error(0, "Unable to locate registry entry for ClearType DLL!"); 
        return -1;
    }
    hLib = LoadLibrary(path);
    if (!hLib) {
        lit_error(ERR_WIN32,"DLL load failed for \"%s\" !", path);
        free(path);
        return -1;
    }
    free(path);

    r = drm5_load_secure_repository();
    return r;
}

int drm5_setkey(U8 * p, int size)
{
    setsymfunc hacksetsym;
    int r, retval;

    retval = 0;
    trampoline_target = (voidfunc)pTable->setsym;
    hacksetsym = (setsymfunc)jumpgate;
    paramcount = 5;
    r = hacksetsym(bbobj, 0,size, p, &key_tag);
    if (r < 0) {
        lit_error(0,"Blackbox \"Set Symmetric Key\" Failed! (%d)\n", r);
        return -1;
    }
    return 0;
}

int drm5_decrypt(U8 * p, int size)
{
    decfunc    hackdec;
    int        result_size, r;

    result_size = size;
    trampoline_target = (voidfunc)pTable->decrypt;
    hackdec = (decfunc)jumpgate;
    paramcount = 6;
    r = hackdec(bbobj, key_tag,size,p,&result_size,p);
    if (r < 0) {
        lit_error(0,"Blackbox \"Decryption\" failed! (%d)\n", r);
        return -1;
    }
    return 0;
}

/*--[drm5_callback]------------------------------------------------------------
 |
 | This routine will handle interfacing to the main lit library.  
 |
 | If p is NULL, then initialize..
 | Otherwise, decrypt and hand back the data.
 | 
*/
int drm5_callback(void * pv, U8 * p, int size)
{
    int status = 0;
    if (!p) {
        status = drm5_init();
        if (!status) status = drm5_handle_key((lit_file *)pv);
    }
    else { status =  drm5_decrypt(p, size); };
    if (status)
    {
        FreeLibrary(hDRM);
        hDRM = NULL;
    }
    return status;
}
